\ "vm-code" and "end-cod" for assembling to the target system


\ tokens build in first-pass

\ content of this token: counted string of a word for evaluating
1 constant word-token

\ content of this token: counted string of a word for evaluating calling
2 constant call-token

\ content of this token: a cell for the origin
3 constant org-token

\ start of the token list
0 value list-start

\ current end of the token list
0 value list-end

\ create a new list
: new-list
    0 to list-start
    0 to list-end
;

\ append one item at the end of the list
: add-item
    list-start 0= if
        here to list-start
    then
    list-end 0<> if
        here list-end !
    then
    here to list-end
    0 ,
;

\ on first call, get the first item of the list, then the rest
\ return 0, if there is no more item
: next-item ( -- addr )
    list-start if
        list-start cell+
        list-start @ 0= if
            0 to list-start
        else
            list-start @ to list-start
        then
    else
        0
    then
;

create next-word-buffer 256 chars allot

\ read the next word from input buffer. Refills the buffer, if empty
\ To avoid invalidation of "," etc., the word is copied to a buffer
: next-word ( -- addr u )
    begin
        bl word count
        dup 0= if
            refill drop 2drop
        else
            >r next-word-buffer r@ cmove
            next-word-buffer r> exit
        then
    again
;

\ allot len bytes and copy len bytes, starting at addr, to the allocated memory
: allot-string ( addr len -- )
    here over allot swap cmove
;

\ read words until right bracket and store all words as one item
: parse-forth ( "...<spaces>]" -- )
    add-item word-token c, 0 c,
    begin
        2dup s" ]" compare 0= if
            2drop
            -1 allot  \ remove last space
            list-end cell+ 1+  \ start of counted string
            here over - 1-  \ length of string
            swap c!
            exit
        else
            allot-string bl c,
        then
    again
;

\ evaluate a word, called by "first-pass"
: evaluate-word ( addr n -- )
    2dup s" :" compare 0= if 2drop vm-here constant exit then
    2dup s" [" compare 0= if 2drop parse-forth 0 , exit then
    2dup s" \" compare 0= if 2drop refill drop exit then
    2dup s" call" compare 0= if
        2drop
        add-item call-token c,
        next-word dup c, allot-string
        0 vm-,
    else
        2dup s" .org" compare 0= if
            2drop
            add-item org-token c,
            next-word 
            evaluate dup , to vm-here
        else
            add-item word-token c,
            dup c, allot-string
            0 vm-,
        then
    then
;

\ create tokens and constants for the source code
: first-pass ( -- )
    new-list
    begin
        next-word
        2dup s" end-code" compare 0= if
            2drop exit
        else
            evaluate-word
        then
    again
;

\ debug output while assembling
: debug-output ( addr u value -- )
	vm-here 1- 4 hex-padded type bl emit
	4 hex-padded type bl emit
	type cr
;

\ now forward references are defined, iterate all tokens and generate code and data
: second-pass
    0 org
    cr
    begin
        next-item
        dup 0= if drop exit then
        dup char+ swap c@ case
            call-token of count 2dup evaluate 8000 or dup vm-, debug-output endof
            word-token of count 2dup evaluate dup vm-, debug-output endof
            org-token of @ dup ." org:" . cr to vm-here endof
        endcase
    again
;

\ assembler for the virtual machine
: vm-code
    first-pass
    second-pass
;
