\ "vm-steps" and "vm-step" for emulating the CPU

0 constant debug-output

\ cpu states
0 constant read-opcode
1 constant step1
2 constant step2

\ initial state
read-opcode value vm-state

\ CPU registers
100 value pc 0 value next-pc
3f value sp 0 value next-sp
ff value rp 0 value next-rp
0 value microcode
0 value next-microcode
0 value a 0 value next-a

\ internal clock counter, for every vm-step
0 value clock

\ RAM control signals
0 value address 0 value next-address
0 value data 0 value next-data
0 value wren 0 value next-wren
0 value q

\ update hardware registers and execute one step of the virtual machine
: vm-step
    \ read/write access
    wren if
        data address vm-!
    else
        address vm-@ to q
    then

    \ test for LED access
    address led-port = wren 1 = and if
        ." LED write: " data . ."  at clock cycle: " clock . cr
sp . cr
rp . cr
    then
    address keyboard-port = if
        ." keyboard read at clock cycle: " clock . cr
    then

    \ save registers for next step
    rp to next-rp
    sp to next-sp
    pc to next-pc
    address to next-address
    data to next-data
    a to next-a
    0 to wren 0 to next-wren

    \ execute step
    0 to microcode
    0 to next-wren
    q to next-data
    vm-state case
        read-opcode of
            pc to next-address
            pc 1+ to next-pc
            step1 to vm-state
            debug-output if ." read opcode " cr then
        endof
        step1 of
            q 8000 and if
                q 7fff and to next-pc
                rp to next-address
                rp 1- to next-rp
                1 to next-wren
                pc to next-data
                read-opcode to vm-state
            else
                q 7f and to microcode
                q 7 rshift 7f and to next-microcode
                step2 to vm-state
            then
        endof
        step2 of
            next-microcode to microcode
            read-opcode to vm-state
        endof
    endcase

    microcode 3 and if
        microcode 78 and case
            m-q2a of q to next-a endof
            m-a2data of a to next-data endof
            m-a2address of a to next-address endof
            m-pc2address of pc to next-address pc 1+ to next-pc endof
            m-pc2data of pc to next-data endof
            m-a2pc of a to next-pc endof
            m-branch of q 0 <> if a to next-pc then endof
            m-q2pc of q to next-pc endof
            m-and of q a and to next-data endof
            m-or of q a or to next-data endof
            m-xor of q a xor to next-data endof
            m-plus of q a + to next-data endof
            m-lshift of q 1 lshift to next-data endof
            m-rshift of q 1 rshift to next-data endof
            m-pop of sp 1+ to next-address sp 1+ to next-sp endof
            m-popr of rp 1+ to next-address rp 1+ to next-rp endof
        endcase
    then

    microcode m-wren and if
        1 to next-wren
    then

    microcode 3 and case
        2 of
            \ push
            sp to next-address
            sp 1- to next-sp
        endof
        3 of
            \ pushr
            rp to next-address
            rp 1- to next-rp
        endof
    endcase

    \ copy changed registers for next step
    next-sp to sp
    next-rp to rp
    next-pc to pc
    next-address to address
    next-data to data
    next-a to a
    next-wren to wren
    debug-output if
        ." registers: a: " a .
        ." sp: " sp .
        ." rp: " rp .
        ." tos: " sp 1+ vm-@ .
        ." tosr: " rp 1+ vm-@ .
        ." nos: " sp 2 + vm-@ .
        ." pc: " pc .
        ." address: " address .
        ." data: " data .
        ." wren: " wren .
        ." q: " q . cr
    then

	\ increment clock counter
	clock 1+ to clock
;

\ execute n steps of the virtual machine
: vm-steps ( n -- )
    0 do vm-step loop
;

: vm-reset
	0 to pc
	read-opcode to vm-state
;
